/*
 * smart_pointer-test-compare.cpp    Smart pointer, Step 3.
 */
#include "smart_pointer.h"
#include <iostream>
#include <iomanip>
using namespace std;
using namespace IDA_Smart_Pointer;

int main()
{
   cout << "Declaring raw pointers with const variations:\n";
   cout << "int*             p   = new int{ 1 };\n";                 
   cout << "const int*       pc  = new int{ 2 };\n";                
   cout << "int* const       cp  = new int{ 3 };\n";      
   cout << "const int* const cpc = new const int{ 4 };\n";
   int*             p   = new int{ 1 };        // pointer to int
   const int*       pc  = new const int{ 2 };  // pointer to constant int
   int* const       cp  = new int{ 3 };        // constant pointer to int
   const int* const cpc = new const int{ 4 };  // constant pointer to constant int
   
   cout << "\nDeclaring corresponding smart pointers:\n";
   cout << "smart_pointer sp1{ p };\n";
   cout << "smart_pointer sp2{ pc }\n";
   cout << "smart_pointer sp3{ cp };\n";
   cout << "smart_pointer sp4{ cpc }\n";
   smart_pointer sp1{ p };
// smart_pointer sp2{ pc };
   smart_pointer sp3{ cp };
// smart_pointer sp4{ cpc };

   cout << "\nDeclaring const smart pointers, initializing with different constructors:\n";
   cout << "const smart_pointer csp0;\n";
   cout << "const smart_pointer csp1{ new int{ 11 } };\n";
   cout << "const smart_pointer csp2{ sp3 };\n";
   const smart_pointer csp0;
   const smart_pointer csp1{ new int{ 11 } };
   const smart_pointer csp2{ sp3 };

   cout << "\nTesting assignment for const smart pointers:\n";
   cout << "csp0 = csp1;  // Not allowed (fine!)\n";
   cout << "csp0 = p;     // Not allowed (fine!)\n";
// csp0 = csp1;  // Not allowed;
// csp0 = p;     // Not allowed;

   cout << "\nComparing const and non-const smart pointers using == and !=\n";
   if (sp1 == csp1) cout << "sp1 == csp1: true\n"; else cout << "sp1 == csp1: false\n";
   if (csp1 == sp1) cout << "csp1 == sp1: true\n"; else cout << "csp1 == sp1: false\n";
   if (sp1 != csp1) cout << "sp1 !=c sp1: true\n"; else cout << "sp1 != csp1: false\n";
   if (csp1 != sp1) cout << "csp1 != sp1: true\n"; else cout << "csp1 != sp1: false\n";

   cout << "\nComparing const smart pointers and raw pointers using == and !=\n";
   if (csp1 == p) cout << "csp1 == p: true\n"; else cout << "csp1 == p: false\n";
   if (p == csp1) cout << "p == csp1: true\n"; else cout << "p == csp1: false\n";
   if (csp1 != p) cout << "csp1 != p: true\n"; else cout << "csp1 != p: false\n";
   if (p != csp1) cout << "p != csp1: true\n"; else cout << "p != csp1: false\n";

   cout << "\nApplying operator-> using ordinary function call syntax on const smart pointers\n";
   cout << "csp0.operator->(): " << csp0.operator->() << '\n';
   cout << "csp1.operator->(): " << csp1.operator->() << '\n';

   cout << "\nApplying operator* using ordinary member functions call syntax on const smart pointers\n";
   cout << "csp1.operator*(): " << csp1.operator*() << '\n';

   // Direct null pointer test is not allowed; there is no type conversion to bool
// if (csp0) cout << "csp0: true\n"; else cout << "csp0: false\n";
// if (csp1) cout << "csp1: true\n"; else cout << "csp1: false\n";

   cout << "\nNegated direct null pointer tests for const null tom pointers\n";
   if (!csp0) cout << "!csp0: true\n"; else cout << "!csp0: false\n";
   if (!csp1) cout << "!csp1: true\n"; else cout << "!csp1: false\n";

   cout << "\nExplicit equality test on const null pointers\n";
   if (csp0 == nullptr) cout << "csp0 == nullptr: true\n"; else cout << "csp0 == nullptr: false\n";
   if (nullptr == csp0) cout << "nullptr == csp0: true\n"; else cout << "nullptr == csp0: false\n";
   if (csp1 == nullptr) cout << "csp1 == nullptr: true\n"; else cout << "csp1 == nullptr: false\n";
   if (nullptr == csp1) cout << "nullptr == csp1: true\n"; else cout << "nullptr == csp1: false\n";

   cout << "\nExplicit inequality tests for const null pointers\n";
   if (csp0 != nullptr) cout << "csp0 != nullptr: true\n"; else cout << "csp0 != nullptr: false\n";
   if (nullptr != csp0) cout << "nullptr != csp0: true\n"; else cout << "nullptr != csp0: false\n";
   if (csp1 != nullptr) cout << "csp1 != nullptr: true\n"; else cout << "csp1 != nullptr: false\n";
   if (nullptr != csp1) cout << "nullptr != csp1: true\n"; else cout << "nullptr != csp1: false\n";

   return 0;
}
